/* Distributed Checksum Clearinghouse
 *
 * routines to make WIN32 look sort of reasonable
 *
 * Copyright (c) 2003 by Rhyolite Software
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.2.31-1.2 $Revision$
 */


#ifdef DCC_WIN32

/* in NT, this should probably have something to do with the event log */

#include "dcc_defs.h"

static HANDLE syslog_fd = INVALID_HANDLE_VALUE;
PATH syslog_nm, syslog_nm2;
PATH default_syslog_nm, default_syslog_nm2;
static char syslog_ident[14+1];
static char syslog_prefix[sizeof(syslog_ident)+12];
static int syslog_mask = LOG_UPTO(LOG_DEBUG);
static int syslog_logopt = LOG_PID | LOG_CONS;
u_char syslog_debug;			/* 1=debug syslog facility */
static time_t lastlog;
static int syslog_failures;
#define MAX_SYSLOG_SIZE (500*1000)


void
get_syslog_default(void)
{
	int i;

	if (default_syslog_nm[0] == '\0') {
		GetTempPath(sizeof(default_syslog_nm),
			    default_syslog_nm);
		i = strlen(default_syslog_nm);
		if (i != 0 && default_syslog_nm[i-1] != '\\')
			default_syslog_nm[i++] = '\\';
		strcpy(default_syslog_nm2, default_syslog_nm);
		strncat(&default_syslog_nm[i], DEFAULT_SYSLOG_NM,
			sizeof(default_syslog_nm)-i-1);
		strncat(&default_syslog_nm2[i], DEFAULT_SYSLOG_NM2,
			sizeof(default_syslog_nm2)-i-1);
	}

	if (syslog_nm[0] == '\0') {
		strcpy(syslog_nm, default_syslog_nm);
		strcpy(syslog_nm2, default_syslog_nm2);
	}
}



#pragma argsused
void
openlog(const char *ident, int logopt, int facility)
{
	PATH buf;

	if (!syslog_guard.label) {
		get_syslog_default();
		syslog_lock_init();
	}

	syslog_logopt = logopt;

	/* use new ident if it is not null, and not just a re-telling
	 * of the current value from inside this library */
	if (ident && *ident != '\0') {
		if (syslog_ident != ident)
			BUFCPY(syslog_ident, ident);
	} else {
		GetModuleNm(0, buf, sizeof(buf));
		ident = strrchr(buf, '\\');
		if (!ident)
			ident = buf;
		else
			ident++;
		BUFCPY(syslog_ident, ident);
	}

	if (syslog_logopt & LOG_PID) {
		sprintf(syslog_prefix, "%s[%x]",
			syslog_ident, GetCurrentProcessId());
	} else {
		sprintf(syslog_prefix, "%s", syslog_ident);
	}

	/* delay creating and opening the file until it is needed */
	closelog();
}



void
closelog(void)
{
	syslog_lock();
	if (syslog_fd != INVALID_HANDLE_VALUE) {
		CloseHandle(syslog_fd);
		syslog_fd = INVALID_HANDLE_VALUE;
	}
	syslog_unlock();
}



u_char					/* 0=not used recently, 1=active */
quietlog(void)
{
	time_t now;

	if (syslog_active)
		return 1;
	if (syslog_fd == INVALID_HANDLE_VALUE)
		return 0;
	now = time(0);
	if (now >= lastlog && now <= lastlog+1)
		return 1;
	closelog();
	return 0;
}



int
setlogmask(int mask)
{
	int omask;

	omask = syslog_mask;
	if (mask != 0)
		syslog_mask = mask;
	return omask;
}



#pragma argsused
static void
vsyslog_fallback(int priority, const char *msg, va_list args)
{
	if (priority <= LOG_CRIT || (syslog_logopt & LOG_PERROR)
	    || syslog_debug)
		vbad_message_box(app_title, -1, msg, args);
}


static void
syslog_fallback(int priority, const char *msg, ...)
{
	va_list args;

	va_start(args, msg);
	vsyslog_fallback(priority, msg, args);
	va_end(args);
}


static u_char				/* 1=ok 0=failed */
syslog_open_file(int priority, DWORD amode, DWORD smode)
{
	int i;

	closelog();

	for (i = 0; i < 15; i++) {
		syslog_fd = CreateFile(syslog_nm, amode, smode,
				       0, OPEN_ALWAYS,
				       FILE_ATTRIBUTE_NORMAL, 0);
		if (syslog_fd != INVALID_HANDLE_VALUE)
			return 1;
		Sleep(10);
	}
	syslog_fallback(priority, "syslog CreateFile(%s): %s",
			syslog_nm, ERROR_STR());
	return 0;
}



/* Is there an equivalent in NT to the old UNIX O_APPEND? */
static u_char				/* 0=failed with message printed */
file_append(HANDLE fd,
	    void *buf,
	    int len,
	    int priority,
	    char *nm)
{
	DWORD d;
	int i, serrno;

	i = 0;
	for (;;) {
		if (LockFile(fd, 0, 0, 1, 0)) {
			if (SetFilePointer(fd, 0, 0, FILE_END) == 0xffffffff) {
				serrno = GetLastError();
				UnlockFile(fd, 0, 0, 1, 0);
				SetLastError(serrno);
				syslog_fallback(priority,
						"SetFilePointer(%s): %s",
						nm, ERROR_STR());
				return 0;
			}
			break;
		}
		if (++i > 10 || GetLastError() != ERROR_LOCK_VIOLATION) {
			syslog_fallback(priority, "LockFile(%s): %s:",
					nm, ERROR_STR());
			return 0;
		}
		Sleep(10);
	}

	if (!WriteFile(fd, buf, len, &d, 0)
	    || d != len) {
		serrno = GetLastError();
		UnlockFile(fd, 0, 0, 1, 0);
		SetLastError(serrno);
		syslog_fallback(priority, "Write(%s,%d)=%d: %s",
				nm, len, d, ERROR_STR());
		return 0;
	}

	UnlockFile(fd, 0, 0, 1, 0);
	return 1;
}



#pragma argsused
void
vsyslog(int priority, const char *msg, va_list args)
{
	time_t now;
	struct tm tm;
	char *bp, buf[MAX_PATH*2+100];	/* just hope it's big enough */
	DWORD sz;
	int i;


	/* do nothing if this priority is turned off */
	if (!(syslog_mask & (1 << (priority & LOG_PRIMASK))))
		return;

	now = time(0);
	if (syslog_failures != 0
	    && now >= lastlog
	    && now <= lastlog+5		/* (guard against time changes) */
	    && priority > LOG_CRIT) {
		/* do not try to re-open the file too often */
		vsyslog_fallback(priority, msg, args);
		syslog_failures++;
		return;
	}
	ts_localtime(&tm, now);

	if (syslog_fd == INVALID_HANDLE_VALUE) {
		openlog(syslog_ident, syslog_logopt, LOG_USER);

		if (syslog_nm[0] == '\0') {
			/* display on the console if necessary */
			vsyslog_fallback(priority, msg, args);
			return;
		}

		if (!syslog_open_file(priority, GENERIC_WRITE | GENERIC_READ,
				      FILE_SHARE_READ | FILE_SHARE_WRITE)) {
			/* display the real message and quit */
			vsyslog_fallback(priority, msg, args);
			syslog_failures++;
			return;
		}

	}

	if (syslog_nm2[0] != '\0'
	    && GetFileSize(syslog_fd, 0) > MAX_SYSLOG_SIZE) {
		/* if the file is big,
		 *	try to open it for reading
		 *	copy it,
		 *	reopen for exclusive writing
		 *	truncate it if no other process has
		 */
		if (!syslog_open_file(priority,
				      GENERIC_READ, FILE_SHARE_READ)) {
			vsyslog_fallback(priority, msg, args);
			syslog_failures++;
			return;
		}
		/* if we lose the race, give up and try again on the
		 * next message */
		sz = GetFileSize(syslog_fd, 0);
		if (sz > MAX_SYSLOG_SIZE
		    && CopyFile(syslog_nm, syslog_nm2, 0)
		    && syslog_open_file(priority, GENERIC_WRITE, 0)
		    && sz == GetFileSize(syslog_fd, 0)
		    && 0 == SetFilePointer(syslog_fd, 0, 0, FILE_BEGIN)
		    && SetEndOfFile(syslog_fd)) {
			i = strftime(buf, sizeof(buf),
				     "%D %H:%M:%S log restarted\r\n", &tm);
			if (!file_append(syslog_fd, buf, i,
					 priority, syslog_nm)) {
				vsyslog_fallback(priority, msg, args);
				closelog();
				syslog_failures++;
				return;
			}
		}

		/* re-open the file to actually use it,
		 * while sharing it with others */
		if (!syslog_open_file(priority, GENERIC_WRITE | GENERIC_READ,
				      FILE_SHARE_READ | FILE_SHARE_WRITE)) {
			vsyslog_fallback(priority, msg, args);
			syslog_failures++;
			return;
		}
	}

	if (syslog_failures != 0) {
		bp = buf+strftime(buf, sizeof(buf), "%D %H:%M:%S ",&tm);
		bp += sprintf(bp, "%d lines lost\r\n", syslog_failures);
		if (!file_append(syslog_fd, buf, bp-buf,
				 priority, syslog_nm)) {
			vsyslog_fallback(priority, msg, args);
			closelog();
			syslog_failures++;
			return;
		}
		syslog_failures = 0;
	}

	strcpy(buf, syslog_prefix);
	bp = buf+strlen(buf);
	bp += strftime(bp, sizeof(buf), " %D %H:%M:%S ", &tm);

	bp += vsnprintf(bp, &buf[sizeof(buf)-3]-bp, msg, args);
	strcpy(bp, "\r\n");
	bp += 2;

	if (!file_append(syslog_fd, buf, bp-buf, priority, syslog_nm)) {
		vsyslog_fallback(priority, msg, args);
		closelog();
		syslog_failures++;
		return;
	}
	lastlog = now;

	/* The line is in the file.  Display it on the console if necessary */
	if (syslog_logopt & LOG_PERROR)
		vsyslog_fallback(LOG_DEBUG, msg, args);
}



void
syslog(int priority, const char *msg, ...)
{
	va_list args;

	va_start(args, msg);
	vsyslog(priority, msg, args);
	va_end(args);
}



/* complete the lame strerror() from Borland for WIN95 */
const char *
ws_strerror(int eno)
{
	static struct {
	    char    s[150];
	} ebufs[8];
	static int ebuf_num;
	int bn;
	const char *src, *end;
	char c, *begin, *tgt;

	/* Use an array of static buffers to kludge around problems with
	 * threads */
	bn = ebuf_num;
	ebuf_num = (bn+1) % DIM(ebufs);
	begin = ebufs[bn].s;

	/* Use the Microsoft message if it is not silly. */
	if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,
			  0, eno, 0, begin, sizeof(ebufs[bn].s), 0)) {

		/* strip any CR or LF and convert the strange, non-ASCII
		 * garbage from of messages */
		src = tgt = begin;
		end = begin+sizeof(ebufs[bn].s)-1;
		do {
			if (tgt >= end) {
				*tgt = '\0';
				break;
			}
			c = *src++;
			if (c == '\r')
				continue;
			if (c == '\n')
				c = ' ';
			if ((c < ' ' && c != '\0') || c > 0x7e)
				c = '?';
			*tgt++ = c;
		} while (c != '\0');
		return begin;
	}

	/* If Microsoft fails, try the Borland messages,
	 * and use anything other than "unknown error" */
	src = strerror(eno);
	if (strcmp(src, "Unknown error\n"))
		return src;

	/* MicroSoft has only some of the BSD standard error messages */
	switch (eno) {
	case WSAEACCES:		return "SO_BROADCAST not enabled";
	case WSAEADDRINUSE:	return "address already in use";
	case WSAEADDRNOTAVAIL:	return "address not available";
	case WSAEAFNOSUPPORT:	return "Address family not supported";
	case WSAEALREADY:	return "nonblocking connect in progress";
	case WSAEBADF:		return "Bad file descriptor";
	case WSAECONNABORTED:	return "Software caused connection abort";
	case WSAECONNREFUSED:	return "Connection refused";
	case WSAECONNRESET:	return "Connection reset by peer";
	case WSAEDESTADDRREQ:	return "Destination address required";
	case WSAEDQUOT:		return "Disc quota exceeded";
	case WSAEFAULT:		return "WS bad address";
	case WSAEHOSTDOWN:	return "Host is down";
	case WSAEHOSTUNREACH:	return "No route to host";
	case WSAEINPROGRESS:	return "winsock 1.1 call in progress";
	case WSAEINTR:		return "cancelled by WSACancelBlockingCall";
	case WSAEINVAL:		return "WS invalid argument";
	case WSAEISCONN:	return "Socket is already connected";
	case WSAELOOP:		return "Too many levels of symbolic links";
	case WSAEMFILE:		return "Too many open files";
	case WSAEMSGSIZE:	return "Message too long";
	case WSAENAMETOOLONG:	return "File name too long";
	case WSAENETDOWN:	return "network is down";
	case WSAENETRESET:	return "Network dropped connection on reset";
	case WSAENETUNREACH:	return "network is unreachable";
	case WSAENOBUFS:	return "No buffer space available";
	case WSAENOPROTOOPT:	return "Protocol not available";
	case WSAENOTCONN:	return "Socket is not connected";
	case WSAENOTEMPTY:	return "Directory not empty";
	case WSAENOTSOCK:	return "socket operation on non-socket";
	case WSAEOPNOTSUPP:	return "Operation not supported";
	case WSAEPFNOSUPPORT:	return "Protocol family not supported";
	case WSAEPROCLIM:	return "Too many processes";
	case WSAEPROTONOSUPPORT:return "Protocol not supported";
	case WSAEPROTOTYPE:	return "Protocol wrong type for socket";
	case WSAEREMOTE:	return "Too many levels of remote in path";
	case WSAESHUTDOWN:	return "Can't send after socket shutdown";
	case WSAESOCKTNOSUPPORT:return "Socket type not supported";
	case WSAESTALE:		return "Stale NFS file handle";
	case WSAETIMEDOUT:	return "Operation timed out";
	case WSAETOOMANYREFS:	return "Too many references: can't splice";
	case WSAEUSERS:		return "Too many users";
	case WSAEWOULDBLOCK:	return "operation would block";
	case WSANOTINITIALISED:	return "WSAStartup not done";
	}

	/* fall back on Borland's "unkonwn error", but admit the error # */
	snprintf(begin, sizeof(ebufs[bn].s), "Unknown error %d", eno);
	return begin;
}
#endif /* !DCC_WIN32 */
